﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

using System.IO;

namespace Engine.KifuWarabe
{
    class Program
    {


        /// <summary>
        /// 作者名です。
        /// </summary>
        private static string authorName = "TAKAHASHI Satoshi";


        #region メイン
        /// <summary>
        /// Ｃ＃のプログラムは、この Main 関数から始まり、 Main 関数の中で終わります。
        /// </summary>
        /// <param name="args"></param>
        static void Main(string[] args)
        {
            #region ↑詳説
            // 
            // 図.
            // 
            //     プログラムの開始：  ここの先頭行から始まります。
            //     プログラムの実行：  この中で、ずっと無限ループし続けています。
            //     プログラムの終了：  この中の最終行を終えたとき、
            //                         または途中で Environment.Exit(0); が呼ばれたときに終わります。
            //                         また、コンソールウィンドウの[×]ボタンを押して強制終了されたときも  ぶつ切り  で突然終わります。
            #endregion

            //-------------------+----------------------------------------------------------------------------------------------------
            // ログファイル削除  |
            //-------------------+----------------------------------------------------------------------------------------------------
            #region ↓詳説
            //
            // 図.
            //
            //      フォルダー
            //          ├─ Engine.KifuWarabe.exe
            //          └─ log.txt               ←これを削除
            //
            #endregion
            try
            {
                System.IO.File.Delete("log.txt");
            }
            catch (Exception)
            {
                // エラーが起こりました。
                //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

                // どうにもできないので  無視します。
            }

            //-------------+----------------------------------------------------------------------------------------------------------
            // ログ書込み  |  ＜この将棋エンジン＞  製品名、バージョン番号
            //-------------+----------------------------------------------------------------------------------------------------------
            #region ↓詳説
            //
            // 図.
            //
            //      log.txt
            //      ┌────────────────────────────────────────
            //      │2014/08/02 1:04:59> v(^▽^)v ｲｪｰｲ☆ ... fugafuga 1.00.0
            //      │
            //      │
            //
            //
            // 製品名とバージョン番号は、次のファイルに書かれているものを使っています。
            // 場所：  [ソリューション エクスプローラー]-[ソリューション名]-[プロジェクト名]-[Properties]-[AssemblyInfo.cs] の中の、[AssemblyProduct]と[AssemblyVersion] を参照。
            //
            // バージョン番号を「1.00.0」形式（メジャー番号.マイナー番号.ビルド番号)で書くのは作者の趣味です。
            //
            #endregion
            string seihinName;
            string versionStr;
            {
                // 製品名
                seihinName = ((System.Reflection.AssemblyProductAttribute)Attribute.GetCustomAttribute(System.Reflection.Assembly.GetExecutingAssembly(), typeof(System.Reflection.AssemblyProductAttribute))).Product;

                // バージョン番号
                Version version = System.Reflection.Assembly.GetExecutingAssembly().GetName().Version;
                versionStr = String.Format("{0}.{1}.{2}", version.Major, version.Minor.ToString("00"), version.Build);
            }
            Program.Log("v(^▽^)v ｲｪｰｲ☆ ... " + seihinName + " " + versionStr);


            //-----------+------------------------------------------------------------------------------------------------------------
            // 通信開始  |
            //-----------+------------------------------------------------------------------------------------------------------------
            #region ↓詳説
            //
            // 図.
            //
            //      無限ループ（全体）
            //          │
            //          ├─無限ループ（１）
            //          │                      将棋エンジンであることが認知されるまで、目で訴え続けます(^▽^)
            //          │                      認知されると、無限ループ（２）に進みます。
            //          │
            //          └─無限ループ（２）
            //                                  対局中、ずっとです。
            //                                  対局が終わると、無限ループ（１）に戻ります。
            //
            // 無限ループの中に、２つの無限ループが入っています。
            //
            #endregion

            //************************************************************************************************************************
            // 無限ループ（全体）
            //************************************************************************************************************************
            while (true)
            {
                //************************************************************************************************************************
                // 無限ループ（１つ目）
                //************************************************************************************************************************
                while (true)
                {
                    try
                    {
                        // 将棋所から何かメッセージが届いていないか、見てみます。
                        string line = Console.In.ReadLine();

                        if (null == line)
                        {
                            // メッセージは届いていませんでした。
                            //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

                            goto gt_NextTime1;
                        }

                        // メッセージが届いています！
                        //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

                        if ("usi" == line)
                        {
                            //------------------------------------------------------------
                            // あなたは USI ですか？
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 1:31:35> usi
                            //      │
                            //
                            //
                            // 将棋所で [対局(G)]-[エンジン管理...]-[追加...] でファイルを選んだときに、
                            // 送られてくる文字が usi です。
                            #endregion
                            Program.Log(line);


                            //------------------------------------------------------------
                            // USI です！！
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:33< id name fugafuga 1.00.0
                            //      │2014/08/02 2:03:33< id author hogehoge
                            //      │2014/08/02 2:03:33< usiok
                            //      │
                            //
                            // プログラム名と、作者名を送り返す必要があります。
                            // オプションも送り返せば、受け取ってくれます。
                            // usi を受け取ってから、5秒以内に usiok を送り返して完了です。
                            #endregion
                            Program.Send("id name "+ seihinName + " " + versionStr);
                            Program.Send("id author "+ Program.authorName);
                            Program.Send("usiok");

                        }
                        else if ("isready" == line)
                        {
                            //------------------------------------------------------------
                            // それでは定刻になりましたので……
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 1:31:35> isready
                            //      │
                            //
                            //
                            // 対局開始前に、将棋所から送られてくる文字が isready です。
                            #endregion
                            Program.Log(line);


                            //------------------------------------------------------------
                            // よろしくお願いします(^▽^)！
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:33< readyok
                            //      │
                            //
                            //
                            // いつでも対局する準備が整っていましたら、 readyok を送り返します。
                            #endregion
                            Program.Send("readyok");
                        }
                        else if ("usinewgame" == line)
                        {
                            //------------------------------------------------------------
                            // 対局時計が ポチッ とされました
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:33> usinewgame
                            //      │
                            //
                            //
                            // 対局が始まったときに送られてくる文字が usinewgame です。
                            #endregion
                            Program.Log(line);


                            // 無限ループ（１つ目）を抜けます。無限ループ（２つ目）に進みます。
                            break;
                        }
                        else if ("quit" == line)
                        {
                            //------------------------------------------------------------
                            // おつかれさまでした
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 1:31:38> quit
                            //      │
                            //
                            //
                            // 将棋エンジンを止めるときに送られてくる文字が quit です。
                            #endregion
                            Program.Log(line);

                            
                            //------------------------------------------------------------
                            // ﾉｼ
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 3:08:34> (^-^)ﾉｼ
                            //      │
                            //
                            //
                            #endregion
                            Program.Log("(^-^)ﾉｼ");

                            // このプログラムを終了します。
                            Environment.Exit(0);
                        }
                        else
                        {
                            //------------------------------------------------------------
                            // ○△□×！？
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // ／(＾×＾)＼
                            //

                            // 通信が届いていますが、このプログラムでは  聞かなかったことにします。
                            // USIプロトコルの独習を進め、対応／未対応を選んでください。
                            //
                            // ログだけ取って、スルーします。
                            #endregion
                            Program.Log(line);
                        }
                    }
                    catch (Exception ex)
                    {
                        // エラーが起こりました。
                        //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

                        // どうにもできないので  ログだけ取って無視します。
                        Program.Log(ex.GetType().Name + " " + ex.Message);
                    }

                gt_NextTime1:
                    ;
                }


                //************************************************************************************************************************
                // 無限ループ（２つ目）
                //************************************************************************************************************************
                int sasitaCount = 0;    // 指した回数

                while (true)
                {
                    try
                    {
                        // 将棋所から何かメッセージが届いていないか、見てみます。
                        string line = Console.In.ReadLine();

                        if (null == line)
                        {
                            // メッセージは届いていませんでした。
                            //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

                            goto gt_NextTime2;
                        }

                        // メッセージが届いています！
                        //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

                        if (line.StartsWith("position"))
                        {
                            //------------------------------------------------------------
                            // これが棋譜です
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:35> position startpos moves 2g2f
                            //      │
                            //
                            // この将棋エンジンは後手で、先手が初手  ▲２六歩  を指されたことが分かります。
                            //


                            // 手番になったときに、“まず”、将棋所から送られてくる文字が position です。
                            // このメッセージを読むと、駒の配置が分かります。
                            //
                            // “が”、まだ指してはいけません。
                            #endregion
                            Program.Log(line);


                            //------------------------------------------------------------
                            // じっとがまん
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │
                            //      │
                            //

                            // 応答は無用です。
                            // 多分、将棋所もまだ準備ができていないのではないでしょうか（？）
                            #endregion

                        }
                        else if (line.StartsWith("go ponder"))
                        {
                            //------------------------------------------------------------
                            // 将棋所が次に呼びかけるまで、考えていてください
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:35> go ponder
                            //      │
                            //

                            // 先読み用です。
                            // 今回のプログラムでは対応しません。
                            //
                            // 将棋エンジンが  将棋所に向かって  「bestmove ★ ponder ★」といったメッセージを送ったとき、
                            // 将棋所は「go ponder」というメッセージを返してくると思います。
                            //
                            // 恐らく  このメッセージを受け取っても、将棋エンジンは気にせず  考え続けていればいいのではないでしょうか。
                            #endregion


                            //------------------------------------------------------------
                            // じっとがまん
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │
                            //      │
                            //

                            // まだ指してはいけません。
                            // 指したら反則です。相手はまだ指していないのだ☆ｗ
                            #endregion
                            Program.Log(line);

                        }
                        else if (line.StartsWith("go mate"))
                        {
                            //------------------------------------------------------------
                            // 詰め将棋を解いてみよ！
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:35> go mate
                            //      │
                            //

                            // 詰め将棋用です。
                            // このソフトでは対応しません。
                            #endregion
                            Program.Log(line);


                            //------------------------------------------------------------
                            // 解けた
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:35< checkmate
                            //      │
                            //
                            // 詰みがあれば checkmate を返せばよいのだと思います。
                            #endregion
                            //Program.Send("checkmate");


                            //------------------------------------------------------------
                            // これは詰まない
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:35< checkmate nomate
                            //      │
                            //
                            // 詰みがなければ checkmate nomate を返返せばよいのだと思います。
                            #endregion
                            //Program.Send("checkmate nomate");


                            //------------------------------------------------------------
                            // 解けなかった☆ｗ
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:35< checkmate timeout
                            //      │
                            //
                            // 時間切れの場合 checkmate timeout を返返せばよいのだと思います。
                            #endregion
                            Program.Send("checkmate timeout");

                        }
                        else if (line.StartsWith("go"))
                        {
                            //------------------------------------------------------------
                            // あなたの手番です
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:36:19> go btime 599000 wtime 600000 byoyomi 60000
                            //      │
                            //

                            // もう指していいときに、将棋所から送られてくる文字が go です。
                            // 「go ponder」「go mate」とは区別します。
                            #endregion
                            Program.Log(line);


                            // ┏━━━━サンプル・プログラム━━━━┓
                            if (0 == sasitaCount)
                            {
                                //------------------------------------------------------------
                                // ２手目、△９四歩
                                //------------------------------------------------------------
                                #region ↓詳説
                                //
                                // 図.
                                //
                                //      log.txt
                                //      ┌────────────────────────────────────────
                                //      ～
                                //      │2014/08/02 2:36:19< bestmove 9c9d
                                //      │
                                //

                                // この将棋エンジンは、後手とします。
                                // ２手目、△９四歩  を決め打ちで返します。
                                #endregion
                                Program.Send("bestmove 9c9d");

                                sasitaCount++;
                            }
                            else if (1 == sasitaCount)
                            {
                                //------------------------------------------------------------
                                // ４手目、△１四歩
                                //------------------------------------------------------------
                                #region ↓詳説
                                //
                                // 図.
                                //
                                //      log.txt
                                //      ┌────────────────────────────────────────
                                //      ～
                                //      │2014/08/02 2:36:20< bestmove 1c1d
                                //      │
                                //

                                // この将棋エンジンは、後手とします。
                                // ４手目、△１四歩  を決め打ちで返します。
                                #endregion
                                Program.Send("bestmove 1c1d");

                                sasitaCount++;
                            }
                            else if (2 == sasitaCount)
                            {
                                //------------------------------------------------------------
                                // ６手目、投了
                                //------------------------------------------------------------
                                #region ↓詳説
                                //
                                // 図.
                                //
                                //      log.txt
                                //      ┌────────────────────────────────────────
                                //      ～
                                //      │2014/08/02 2:36:21< bestmove resign
                                //      │
                                //

                                // この将棋エンジンは、後手とします。
                                // ６手目、投了  を決め打ちで返します。
                                #endregion
                                Program.Send("bestmove resign");

                                sasitaCount++;
                            }
                            // ┗━━━━サンプル・プログラム━━━━┛

                        }
                        else if (line.StartsWith("stop"))
                        {
                            //------------------------------------------------------------
                            // あなたの手番です  （すぐ指してください！）
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:03:35> stop
                            //      │
                            //

                            // すぐ指してほしいときに、将棋所から送られてくる文字が stop です。
                            // 「急いで指すボタン」が押されたときなどに送られてくるようです？
                            //
                            // stop するのは思考です。stop を受け取ったら最善手を指してください。
                            #endregion
                            Program.Log(line);


                            //------------------------------------------------------------
                            // じゃあ、△９二香 で
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:36:21< bestmove 9a9b
                            //      │
                            //

                            //
                            // （場合１）特に何もない場合、送ったメッセージ通り、指されます。
                            //
                            // （場合２）１手前の指し手で、将棋エンジンが「bestmove ★ ponder ★」という形で  予想手付きで将棋所にメッセージを送っていたとき、
                            //           その予想手が外れていたならば、将棋所は「stop」を返してきます。
                            //           このとき  思考を打ち切って最善手の指し手をすぐに返信するわけですが、将棋所はこの返信を無視します☆ｗ
                            //           （この指し手は、外れていた予想手について考えていた“最善手”ですからゴミのように捨てられます）
                            //           その後、将棋所から「position」「go」が再送されてくるのだと思います。
                            //
                            //             将棋エンジン「bestmove ★ ponder ★」
                            //                 ↓
                            //             将棋所      「stop」
                            //                 ↓
                            //             将棋エンジン「うその指し手返信」（無視されます）←今ここ
                            //                 ↓
                            //             将棋所      「position」「go」
                            //                 ↓
                            //             将棋エンジン「本当の指し手」
                            //
                            //           という流れと思います。
                            #endregion
                            Program.Send("bestmove 9a9b");

                        }
                        else if (line.StartsWith("ponderhit"))
                        {
                            //------------------------------------------------------------
                            // あなたの手番です  （予想通りの応手でしたよ）
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 2:36:19> ponderhit
                            //      │
                            //

                            // 先読み用です。
                            // 今回のプログラムでは対応しません。
                            //
                            // １手前に指した手のとき、
                            // 将棋エンジンが  将棋所に向かって  「bestmove ★ ponder ★」という形でメッセージを送っていたとき、
                            //
                            // 将棋所は  相手が予想した手を指した後で「go」の替わりに「ponderhit」というメッセージを返してくるのだと思います。
                            // このあとの流れは go と同じだと思います。
                            #endregion
                            Program.Log(line);

                            // 投了ｗ！
                            Program.Send("bestmove resign");

                        }
                        else if (line.StartsWith("gameover"))
                        {
                            //------------------------------------------------------------
                            // 対局が終わりました
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // 図.
                            //
                            //      log.txt
                            //      ┌────────────────────────────────────────
                            //      ～
                            //      │2014/08/02 3:08:34> gameover lose
                            //      │
                            //

                            // 対局が終わったときに送られてくる文字が gameover です。
                            #endregion
                            Program.Log(line);


                            // 無限ループ（２つ目）を抜けます。無限ループ（１つ目）に戻ります。
                            break;
                        }
                        else
                        {
                            //------------------------------------------------------------
                            // ○△□×！？
                            //------------------------------------------------------------
                            #region ↓詳説
                            //
                            // ／(＾×＾)＼
                            //

                            // 通信が届いていますが、このプログラムでは  聞かなかったことにします。
                            // USIプロトコルの独習を進め、対応／未対応を選んでください。
                            //
                            // ログだけ取って、スルーします。
                            #endregion
                            Program.Log(line);
                        }
                    }
                    catch (Exception ex)
                    {
                        // エラーが起こりました。
                        //>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

                        // どうにもできないので  ログだけ取って無視します。
                        Program.Log(ex.GetType().Name + " " + ex.Message);
                    }

                gt_NextTime2:
                    ;
                }
            }
        }
        #endregion

        /// <summary>
        /// ログ・ファイルに通信を記録します。
        /// </summary>
        /// <param name="line"></param>
        static void Log(string line)
        {
            // ログ追記
            System.IO.File.AppendAllText("log.txt", DateTime.UtcNow.ToString() + "> " + line + Environment.NewLine);
        }

        /// <summary>
        /// 将棋所に向かってメッセージを送り返すとともに、
        /// ログ・ファイルに通信を記録します。
        /// </summary>
        /// <param name="line"></param>
        static void Send(string line)
        {
            // 将棋所に向かって、文字を送り返します。
            Console.Out.WriteLine(line);

            // ログ追記
            System.IO.File.AppendAllText("log.txt", DateTime.UtcNow.ToString() + "< " + line + Environment.NewLine);
        }

    }
}
